@@ -7,7 +7,8 @@ from datetime import datetime
 from types import SimpleNamespace
 from urllib.parse import urlencode
 from ._api_object import ApiObject
-from ._util import encode_data_uri, file_to_bytes
+from ._size import human_read_to_bytes
+from ._util import encode_data_uri, file_to_bytes, file_to_chunks, bytes_to_chunks
 from .error import Timeout
 from .models import Model
 
@@ -15,7 +16,7 @@ from .models import Model
 class Jobs:
     """The `Jobs` object.
 
-    This object is used to retreive information about jobs from the API.
+    This object is used to retrieve information about jobs from the API.
 
     Note:
         This class should not be instantiated directly but rather accessed through the `jobs`
@@ -30,7 +31,7 @@ class Jobs:
         IN_PROGRESS='IN_PROGRESS',
         COMPLETED='COMPLETED',
         CANCELED='CANCELED',
-        TIMEOUT='TIMEOUT',
+        TIMEDOUT='TIMEDOUT',
     )
     """Possible job statuses."""
 
@@ -68,8 +69,8 @@ class Jobs:
         Args:
             user (Optional[str]): Name of the job submitter
             access_key (Optional[str]): Identifier of the access key to be assigned to the user
-            start_date (Optional[datetime, str]): initial date to filter recods
-            end_date (Optional[datetime, str]): final date to filter recods
+            start_date (Optional[datetime, str]): initial date to filter records
+            end_date (Optional[datetime, str]): final date to filter records
             model (Optional[str]): Model name or version identifier
             status (Optional[str]): Status of the jobs (all, pending, terminated)
             sort_by (Optional[str]): attribute name to sort results
@@ -183,13 +184,14 @@ class Jobs:
                 raise Timeout('timed out before completion')
         # TODO: should probably ramp up poll_interval as wait time increases
 
-    def submit_text(self, model, version, source, source_name='job'):
+    def submit_text(self, model, version, source, explain=False, source_name='job'):
         """Submits text data for a single source `Job`.
 
         Args:
             model (Union[str, Model]): The model identifier or a `Model` instance.
             version (str): The model version string.
             source (dict): A mapping of model input filename to text string.
+            explain (bool): indicates if you desire an explainable result for your model.`
             source_name (Optional[str]): The source name. Defaults to 'job'.
 
         Returns:
@@ -209,9 +211,9 @@ class Jobs:
                     })
         """
         sources = {source_name: source}
-        return self.submit_text_bulk(model, version, sources)
+        return self.submit_text_bulk(model, version, sources, explain)
 
-    def submit_text_bulk(self, model, version, sources):
+    def submit_text_bulk(self, model, version, sources, explain=False):
         """Submits text data for a multiple source `Job`.
 
         Args:
@@ -219,6 +221,7 @@ class Jobs:
             version (str): The model version string.
             sources (dict): A mapping of source names to text sources. Each source should be a
                 mapping of model input filename to text string.
+            explain (bool): indicates if you desire an explainable result for your model.`
 
         Returns:
             Job: The submitted `Job` instance.
@@ -251,6 +254,7 @@ class Jobs:
                 "identifier": identifier,
                 "version": version
             },
+            "explain": explain,
             "input": {
                 "type": "text",
                 "sources": sources
@@ -260,13 +264,14 @@ class Jobs:
         response = self._api_client.http.post(self._base_route, body)
         return Job(response, self._api_client)
 
-    def submit_bytes(self, model, version, source, source_name='job'):
+    def submit_bytes(self, model, version, source, explain=False, source_name='job'):
         """Submits bytes-like data for a single source `Job`.
 
         Args:
             model (Union[str, Model]): The model identifier or a `Model` instance.
             version (str): The model version string.
             source (dict): A mapping of model input filename to bytes-like object.
+            explain (bool): indicates if you desire an explainable result for your model.`
             source_name (Optional[str]): The source name. Defaults to 'job'.
 
         Returns:
@@ -286,9 +291,9 @@ class Jobs:
                 })
         """
         sources = {source_name: source}
-        return self.submit_bytes_bulk(model, version, sources)
+        return self.submit_bytes_bulk(model, version, sources, explain)
 
-    def submit_bytes_bulk(self, model, version, sources):
+    def submit_bytes_bulk(self, model, version, sources, explain=False):
         """Submits bytes-like data for a multiple source `Job`.
 
         Args:
@@ -296,6 +301,7 @@ class Jobs:
             version (str): The model version string.
             sources (dict): A mapping of source names to text sources. Each source should be a
                 mapping of model input filename to bytes-like object.
+            explain (bool): indicates if you desire an explainable result for your model.`
 
         Returns:
             Job: The submitted `Job` instance.
@@ -334,6 +340,7 @@ class Jobs:
                 "identifier": identifier,
                 "version": version
             },
+            "explain": explain,
             "input": {
                 "type": "embedded",
                 "sources": sources
@@ -343,13 +350,14 @@ class Jobs:
         response = self._api_client.http.post(self._base_route, body)
         return Job(response, self._api_client)
 
-    def submit_files(self, model, version, source, source_name='job'):
+    def submit_files(self, model, version, source, explain=False, source_name='job'):
         """Submits filepath or file-like data for a single source `Job`.
 
         Args:
             model (Union[str, Model]): The model identifier or a `Model` instance.
             version (str): The model version string.
             source (dict): A mapping of model input filename to filepath or file-like object.
+            explain (bool): indicates if you desire an explainable result for your model.`
             source_name (Optional[str]): The source name. Defaults to 'job'.
 
         Returns:
@@ -371,9 +379,9 @@ class Jobs:
                 })
         """
         sources = {source_name: source}
-        return self.submit_files_bulk(model, version, sources)
+        return self.submit_files_bulk(model, version, sources, explain)
 
-    def submit_files_bulk(self, model, version, sources):
+    def submit_files_bulk(self, model, version, sources, explain=False):
         """Submits filepath or file-like data data for a multiple source `Job`.
 
         Args:
@@ -381,6 +389,7 @@ class Jobs:
             version (str): The model version string.
             sources (dict): A mapping of source names to text sources. Each source should be a
                 mapping of model input filename to filepath or file-like object.
+            explain (bool): indicates if you desire an explainable result for your model.`
 
         Returns:
             Job: The submitted `Job` instance.
@@ -404,16 +413,63 @@ class Jobs:
                         }
                     })
         """
-        sources = {
-            source: {
-                key: file_to_bytes(value)
-                for key, value in inputs.items()
-            }
-            for source, inputs in sources.items()
+        identifier = Model._coerce_identifier(model)
+        version = str(version)
+        body = {
+            "model": {
+                "identifier": identifier,
+                "version": version
+            },
+            "explain": explain
         }
-        return self.submit_bytes_bulk(model, version, sources)
+        # Open the job with an empty call to the job api
+        open_job = Job(self._api_client.http.post(self._base_route, body), self._api_client)
+        self.logger.debug("open job %s", open_job)
+        chunk_size = None
+        try:
+            job_features = self.get_features()
+            chunk_size = human_read_to_bytes(job_features["input_chunk_maximum_size"])
+        except:
+            self.logger.warning("Error getting features, assuming defaults")
+            chunk_size = 1024*1024
 
-    def submit_aws_s3(self, model, version, source, access_key_id, secret_access_key, region, source_name='job'):
+        try:
+            # Iterate on the sources, submitting each input as a multipart post request
+            # jobIdentifier/input-item-name/model-input-name
+            for source, inputs in sources.items():
+                for key, value in inputs.items():
+                    self._append_input(open_job, source, key, value, chunk_size)
+
+            open_job = self._api_client.http.post('{}/{}/close'.format(self._base_route, open_job.job_identifier))
+            self.logger.debug("close job %s", open_job)
+        except:
+            try:
+                # Try to cancel the job as something unexpected happened, ignore any error if something bad happen
+                # with this call in order to pass the real cause to the caller
+                self.logger.debug("canceling job %s", open_job)
+                open_job.cancel()
+            finally:
+                pass
+            raise
+        return Job(open_job, self._api_client)
+
+    def _append_input(self, job, input_item_name, data_item_name, input_value, chunk_size):
+        iterable = None
+        if isinstance(input_value, (bytes, bytearray)):
+            iterable = bytes_to_chunks(input_value, chunk_size)
+        else:
+            iterable = file_to_chunks(input_value, chunk_size)            
+        
+        for i, chunk in enumerate(iterable):            
+            self.logger.debug("_append_input(%s, %s, %s, %s, %s) chunk %i", job.job_identifier, input_item_name, data_item_name, type(input_value), chunk_size, i )
+            self._api_client.http.post(
+                '{}/{}/{}/{}'.format(self._base_route, job.job_identifier, input_item_name, data_item_name),
+                None,
+                {"input": chunk}
+            )            
+
+    def submit_aws_s3(self, model, version, source, access_key_id, secret_access_key, region, explain=False,
+                      source_name='job'):
         """Submits AwS S3 hosted data for a single source `Job`.
 
         Args:
@@ -423,6 +479,7 @@ class Jobs:
             access_key_id (str): The AWS Access Key ID.
             secret_access_key (str): The AWS Secret Access Key.
             region (str): The AWS Region.
+            explain (bool): indicates if you desire an explainable result for your model.`
             source_name (Optional[str]): The source name. Defaults to 'job'.
 
         Returns:
@@ -452,9 +509,9 @@ class Jobs:
                 )
         """
         sources = {source_name: source}
-        return self.submit_aws_s3_bulk(model, version, sources, access_key_id, secret_access_key, region)
+        return self.submit_aws_s3_bulk(model, version, sources, access_key_id, secret_access_key, region, explain)
 
-    def submit_aws_s3_bulk(self, model, version, sources, access_key_id, secret_access_key, region):
+    def submit_aws_s3_bulk(self, model, version, sources, access_key_id, secret_access_key, region, explain=False):
         """Submits AwS S3 hosted data for a multiple source `Job`.
 
         Args:
@@ -465,6 +522,7 @@ class Jobs:
             access_key_id (str): The AWS Access Key ID.
             secret_access_key (str): The AWS Secret Access Key.
             region (str): The AWS Region.
+            explain (bool): indicates if you desire an explainable result for your model.`
 
         Returns:
             Job: The submitted `Job` instance.
@@ -515,6 +573,7 @@ class Jobs:
                 "identifier": identifier,
                 "version": version
             },
+            "explain": explain,
             "input": {
                 "type": "aws-s3",
                 "accessKeyID": access_key_id,
@@ -527,7 +586,7 @@ class Jobs:
         response = self._api_client.http.post(self._base_route, body)
         return Job(response, self._api_client)
 
-    def submit_jdbc(self, model, version, url, username, password, driver, query):
+    def submit_jdbc(self, model, version, url, username, password, driver, query, explain=False):
         """Submits jdbc query as input for a `Job`, each row is interpreted as a input.
             Modzy will create a data source with the parameters provided and will execute
             the query provided, then will match the inputs defined of the model with the columns
@@ -541,6 +600,7 @@ class Jobs:
             password (str): the password to access the database
             driver (str): full driver class name
             query (str): the query to execute
+            explain (bool): indicates if you desire an explainable result for your model.`
 
         Returns:
             Job: The submitted `Job` instance.
@@ -577,6 +637,7 @@ class Jobs:
                 "identifier": identifier,
                 "version": version
             },
+            "explain": explain,
             "input": {
                 "type": "jdbc",
                 "url": url,
@@ -590,6 +651,21 @@ class Jobs:
         response = self._api_client.http.post(self._base_route, body)
         return Job(response, self._api_client)
 
+    def get_features(self):
+        """Gets the `Job` features
+
+        Returns:
+            ApiObject: a generic api object with the feature name as keys.
+
+        Raises:
+            ApiError: A subclass of ApiError will be raised if the API returns an error status,
+                or the client is unable to connect.
+        """
+        self.logger.debug("getting features ")
+
+        json_obj = self._api_client.http.get('{}/features'.format(self._base_route))
+        return ApiObject(json_obj, self._api_client)
+
 
 class Job(ApiObject):
     """A job object.
@@ -608,6 +684,7 @@ class Job(ApiObject):
         key name (``job.job_identifier`` or ``job.jobIdentifier``). Alternatively, the original
         "camelCase" JSON key can be used with bracketed key access notation (``job['jobIdentifier']``).
     """
+
     def __init__(self, json_obj, api_client=None):
         if 'status' not in json_obj:
             json_obj['status'] = Jobs.status.SUBMITTED
